'''
Resistivity in Less Than Three Dimensions - Resistivity calculations
University of Toronto. Fall 2010.
Written by Travis Durjan. Original by Carol McClary.
Written for Python 2.6.6, numpy 1.5.1, matplotlib 1.0.0, scipy 0.8.0.
----------
This program calculates resistivities and prints the average resistivity,
standard deviation, and standard error.
Optionally prints resistivity values and their errors in a tabular sequence.
Optionally plots resistivity vs distance (of P-probes).

Can be used to process single files or an entire directory. The latter
necessitates a naming convention for files. See toolpy.parsefilename for
the proposed convention. If that is inadequate, toolpy.parsefilename can
be modified appropriately without affecting this program, as long as the
order, type and number of returned variables are preserved.

**It is highly recommended that a naming convention be used.**

This program does assume that offset data files use the same convention
where current is set at 0.0 (as is the offset data by definition). For
calculations, the mean of the offset is used to obtain corrected voltages.

Position values in the .txt file need not be specified at every point.
For example, if the following were the first two lines in a data file,
0	5	5	5	2.4	5	2.6	5	0.3545
				2.2		2.8		0.4565
it would be interpreted as
0	5	5	5	2.4	5	2.6	5	0.3545
0	5	5	5	2.2	5	2.8	5	0.4565
Also,
		5		2.4		2.6		0.3545
1		4						0.4565
is interpreted as
0	0	5	0	2.4	0	2.6	0	0.3545
1	0	4	0	2.4	0	4.6	0	0.4565
In words, completely missing values are assumed to be zero, and the last
specified value is carried forward until a new value is found. This applies
to all columns, including the voltage column.
For more on the parsing routines, see toolpy.readdata.
'''

# constants
foilwidth 	= .00229	#Enter the width of your foil strip (in m)
foilwidtherr	= .00010	#Enter the uncertainty of the width of your foil (in m)
foildepth 	= .000021	#Enter the thickness of your foil strip (in m)
foildeptherr	= .000002	#Enter the uncertainty of the thickness of your foil strip (in m)
pointererr	= .00030	#Enter the uncertainty of the position any one probe (in m)

# File paths for data and offset data.
# If datapath is a directory, all files will be read and parsed based on naming scheme.
# If a file is specified, only that file is processed.
datapath = 'oct15 - 1d'
#datapath = 'oct26 - 3d\\3D-W-1.0 01.txt'

# If offsetfile cannot be found based on the filename of datafile, use below path.
# This can be used as a "master" offset for a set of similar runs.
offsetpath = 'oct26 - 3d\\3D-W-0.0 01.txt'

# if dim and current cannot be obtained from filename, these values will be used
dimension	= 2	# dimension {1,2,3}
current 	= 0.5	# current value used (in A)

# switches (unless stated otherwise, switches are True/False)
printvalues = False # prints table of b-distance|Resistivity|Error|Voltage
plot = True 	# plots of resistivity vs b-distance (as points)
sameplot = True	# if True, plots are overlayed, else a new window is created per plot
savefig = False	# saves plot(s) as a Portable Network Graphic (.png) in a folder called 'pics'
scale = 'linear' # 'linear' | 'log | 'symlog' - the scale of x/y axes for the plot

fit = 'spline',3 # 'linear' | 'poly' | 'spline' | 'line' | 'none'
# 'linear' - a best fit line (polynomial of degree 1)
# 'poly' - a polynomial fit with the specified number of degrees
# 'spline' - data is interpolated and plotted using splines, see splinefit() for more details
# 'line' - points are simply joined by straight lines
# 'none' (or other) - no lines are drawn


# global imports
import toolpy as tp
from os import path
if plot: import matplotlib.pyplot as plt

# Calculates resistivities and their errors of the specified data set
def resiscalc (data,offsetdata,dim,I,method=None,id=None):
	from numpy import empty, sqrt, sum

	# create a 2-dim empty array for b-distances, resistivities and their errors
	resistivities = empty((len(data[:,0]),3))

	# calculate the error of b (or a), i.e. the error in any inter-probe distance
	berr = sqrt(2*(pointererr**2))

	# error for any calculated voltage (from the standard deviation of offset,
	# using 1 degree of freedom, and converting to V from mV).
	Verr = (sqrt(2*(offsetdata[:,-1].std(ddof=1)**2)))/1000

	# calculate voltages, b-distances and resistivities (dimension-dependent)
	offset = offsetdata[:,-1].mean() # average of offset voltages
	V = (data[:,-1]-offset)/1000 # corrected voltages, converted to V
	
	# calculate distances by Euclidean geometry (allows for skewed positioning of probes in dim>1)
	resistivities[:,0] = b = tp.getb(data)

	if dim == 1: # 1-dimension calculations
		# calculate the cross section area of the foil and its uncertainty
		A = foilwidth*foildepth
		Aerr = A*sqrt((foilwidtherr/foilwidth)**2 + (foildeptherr/foildepth)**2)
		
		resistivities[:,1] = (A*V)/(b*I)
		resistivities[:,2] = resistivities[:,1]*sqrt((Verr/V)**2 + (Aerr/A)**2 + (berr/b)**2)

	elif dim == 2: # 2-dimension calculations
		from numpy import pi
		a = tp.geta(data)
		resistivities[:,1] = (2*pi*a*foildepth*V)/(b*I)
		resistivities[:,2] = resistivities[:,1]*sqrt((Verr/V)**2 + (foildeptherr/foildepth)**2 +  (berr/a)**2 + (berr/b)**2)
	
	elif dim == 3: # 3-dimension calculations
		from numpy import pi
		a = tp.geta(data)
		if sum(a) != sum(b): raise NameError('\'{0}D {1} at {2}A  id:{3}\' is not in a Wenner arrangement.'.format(dim,tp.getmethod(method),I,id))
		resistivities[:,1] = (2*pi*b*V)/I
		resistivities[:,2] = resistivities[:,1]*sqrt((Verr/V)**2 + (berr/b)**2)
	
	else: raise ErrorName('dimension =',dim)
	
	# values to always print
	print '{0}D {1} at {2}A  id:{3}'.format(dim,method,I,id)
	print '    resistivity:\t',resistivities[:,1].mean() # avg of resistivities
	print '    std:\t\t',resistivities[:,1].std(ddof=1) # std w/ 1 degree of freedom
	print '    err:\t\t',sqrt(sum(resistivities[:,2]**2))/len(resistivities[:,2]) # RSS? Std? Some mixture?
	
	# print table with b-distance, resistivities, and voltages
	if printvalues:
		from numpy import arange
		print 'b (cm) Resistivity\t  +/- Error\t\tVoltage (mV)'
		for i in arange(len(b)): print '{0}\t{1:.5e}\t{2:.5e}\t{3:.5f}'\
			.format(b[i]*100,resistivities[i,1],resistivities[i,2],V[i]*1000)
	print ''
	
	# numpy.ndarray with 3 columns: b|resistivity|resis_err
	return resistivities

# Interpolate and plot data. Note: interpolation requires a great deal of processing and
# can be very slow with large data sets or slower systems.
def splinefit (resis,label=None):
	from numpy import arange, argsort, array, linspace, where
	# Multiple b-distances in the data throws off interpolation and so all resistivities
	# at any single b are averaged and this stripped down array is interpolated instead.
	# Data is also sorted first to simplify prep work and allows for analysis by multiple
	# interpolation methods.
	resis_sort = resis.take(argsort(resis[:,0]),axis=0) # sorted by b column
	bs,rs,btemp = [],[],0
	for i in arange(len(resis_sort[:,0])):
		if btemp != resis_sort[i,0]: # if new b found
			btemp = resis_sort[i,0] # set temp to new b
			bs.append(btemp) # save new b
			# find indices where new b occurs and save avg of resis at those indices
			indices = where(resis_sort[:,0]==btemp)
			rs.append(array([resis_sort[i,1] for i in indices]).mean())
	bs,rs = array(bs),array(rs) # convert python lists to numpy.ndarrays
	
	# Can use interp1d or UnivariateSpline or ... 
	# I suggest looking at the documentation for details on how to use these functions.
	#from scipy.interpolate import interp1d
	#f = interp1d(bs,rs,kind='cubic') # kind = {'linear','quadratic','cubic',..}
	from scipy.interpolate import UnivariateSpline
	f = UnivariateSpline(bs,rs,k=3,s=1) # k = {1,2,..,5}, s = float (smoothing factor)
	
	x = linspace(bs.min(),bs.max(),100)
	plt.plot(x,f(x),'-',label=label)
	
# Polynomial plot
def polyfit (resis,label=None,degree=1):
	from numpy import polyfit, poly1d, linspace
	f = poly1d(polyfit(resis[:,0],resis[:,1],degree))
	x = linspace(resis[:,0].min(),resis[:,0].max(),100)
	plt.plot(x,f(x),'-',label=label)

# Plots resistivity vs b-distance. Some properties of the plot can be changed by
# the switches at the beginning of this file.
def resisplot (resis,dim=None,I=None,method=None,id=None):
	if not sameplot: # then create a new figure per plot
		if scale == 'log': plt.figure(figsize=(8,8),dpi=80)
		else: plt.figure()
	
	# plot points with error bars
	plt.errorbar(resis[:,0],resis[:,1],yerr=resis[:,2],fmt='r.')
	
	plt.xscale(scale)
	plt.yscale(scale)
	if method == 'Wenner' and not sameplot: plt.xlabel('a-distance (m)')
	else: plt.xlabel('b-distance (m)')
	plt.ylabel(r'Resistivity ($\Omega\cdot$m)')
	
	# try to use scientific notation (requires matplotlib 1.0+)
	try: plt.ticklabel_format(style='scientific',axis='both')
	except AttributeError: sys.exc_clear()
	
	label = '{1} at {2}A  id:{3}'.format(dim,method,I,id)
	
	# data fitting
	if fit[0] == 'linear':  polyfit(resis,label,1) # polynomial fit of degree 1
	elif fit[0] == 'poly': polyfit(resis,label,fit[1]) # polynomial fit of specified degree
	elif fit[0] == 'spline': splinefit(resis,label) # interpolation and spline
	elif fit[0] == 'line': plt.plot(resis[:,0],resis[:,1],'-',label=label)
	
	if sameplot: 
		plt.title('{0}D Resitivity vs b'.format(dim))
		plt.legend(loc='best')
		if savefig: figname = '{0}D Resitivity overlay {1}'.format(dim,scale)
	else:
		title = '{0}D {1} at {2}A  id:{3}'.format(dim,method,I,id)
		plt.title(title)
		if savefig: figname = title.replace(':','-')+scale
	
	if savefig: plt.savefig('{0}\\{1}.png'.format(tp.createdir('pics'),figname),\
		bbox_inches='tight',dpi=80)
	
# This essentially "puts everything together". Performs all necessary actions on the
# specified file. If file is already a full path, simply set directory='' (an empty string).
def processfile (directory,file):
	dim,meth,I,id = tp.parsefilename(path.basename(file))
	if dim == None: dim = dimension
	if I == None: I = current
	
	if I != 0: # skip offset data files
		offsetfile = file.replace(str(I),'0.0')
		if path.exists(path.join(directory, offsetfile)):
			data = tp.readdata(path.join(directory,file))
			offsetdata = tp.readdata(path.join(directory,offsetfile))
			resis = resiscalc(data,offsetdata,dim,I,meth,id)
			if plot: resisplot(resis,dim,I,meth,id)
		
		elif path.exists(offsetpath):
			data = tp.readdata(path.join(directory,file))
			offsetdata = tp.readdata(offsetpath)
			resis = resiscalc(data,offsetdata,dim,I,meth,id)
			if plot: resisplot(resis,dim,I,meth,id)
		
		else:	print file,'does not have an offset data file.'

# ensure log plots have 1:1 axes
if plot and sameplot and scale == 'log': plt.figure(figsize=(8,8),dpi=80)

# if directory specified, process all files in directory (will not look in sub-directories)
if path.isdir(datapath):
	from os import listdir
	[processfile(datapath,file) for file in listdir(datapath) \
		if path.isfile(path.join(datapath, file))]

# if data file is specified, process just that file
elif path.isfile(datapath): processfile('',datapath)
else: print 'No files found at path:',datapath

if plot: plt.show()
